from logos_cdi.abstract import AbstractContext


class LogosException(Exception):

    def __init__(self, message: str, context: AbstractContext):
        super().__init__(message)
        self.context = context


class ResourceNotFoundException(LogosException):

    def __init__(self, resource_name: str, context: AbstractContext):
        super().__init__(f'resource "{resource_name}" not found', context)
        self.resource_name = resource_name



from abc import ABC, abstractmethod
from argparse import ArgumentParser
from typing import Optional, Mapping
from logos_cdi.abstract import AbstractReference, AbstractResource
from logos_cdi.application import AbstractModule, ModuleMixin
from logos_cdi.container import Container
from logos_cdi.resource import Service, Parameter, Singleton
from logos_cdi.reference import Reference, Parent, AllRegisteredReference


class AbstractCommand(ABC):

    def __init__(self, argument_parser: Optional[ArgumentParser] = None):
        self.argument_parser = argument_parser
    
    def define_arguments(self, argument_parser: ArgumentParser):
        self.argument_parser = argument_parser
    
    @property
    def arguments(self):
        return self.argument_parser.parse_known_args()[0]
    
    @abstractmethod
    def execute(self):
        ...


class DelegateCommand(AbstractCommand):

    def __init__(self, argument_parser: ArgumentParser, commands: Mapping[str, AbstractCommand]):
        super().__init__(argument_parser)
        self.commands = commands
        self.define_arguments(argument_parser)
    
    def define_arguments(self, argument_parser):
        parsers = argument_parser.add_subparsers(dest='command', required=True)
        for command_name, command in self.commands.items():
            command_parser: ArgumentParser = parsers.add_parser(command_name)
            command.define_arguments(command_parser)
        super().define_arguments(argument_parser)
    
    def execute(self):
        return self.commands[self.arguments.command].execute()

class Commands(AbstractModule):

    def define_context(self, context):
        return context.nested(
            Container({
                'commands.delegate_command.class': Parameter('logos_cdi.command.DelegateCommand'),
                'commands.registry': Singleton(
                    class_path='logos_cdi.registry.Registry',
                    parameters={
                        'parent': Parent('commands.registry'),
                        'type': AbstractReference
                    }
                ),
                'commands.argument_parser': Parameter(ArgumentParser()),
                'commands.command': Service(
                    class_path=Reference('commands.delegate_command.class'),
                    parameters={
                        'argument_parser': Reference('commands.argument_parser'),
                        'commands': AllRegisteredReference('commands.registry')
                    }
                )
            })
        )


class CommandModule(ModuleMixin):

    command_prefix = 'commands'

    @abstractmethod
    def define_commands(self) -> Mapping[str, AbstractResource]:
        ...

    def define_context(self, context):
        if context.has('commands.registry'):
            commands = self.define_commands()
            context = context.nested(Container({
                f'{self.command_prefix}.{name}': resource
                for name, resource in commands.items()
            }))

            registry = context.get('commands.registry')
            for name in commands.keys():
                registry.register(name, Reference(f'{self.command_prefix}.{name}'))
        return super().define_context(context)



def main():
    from logos_cdi.application import Application, AbstractModule
    from logos_cdi.container import Container
    from os import environ, getcwd
    from sys import path
    from importlib import import_module

    class ManagementCommand(AbstractCommand):

        def define_arguments(self, argument_parser):
            argument_parser.add_argument('--app-url', default=environ.get('LOGOS_APPLICATION_URL', 'main:app'))
            return super().define_arguments(argument_parser)
        
        def execute(self):
            path.append(getcwd())
            self.define_arguments(self.argument_parser)
            
            module_path, app_var = self.arguments.app_url.split(':')
            app: Application = getattr(import_module(module_path), app_var)

            has_event_emitter = app.has('events.emitter') 

            if has_event_emitter:
                app.get('events.emitter').emit('app:start', {})

            app.extends([
                ManagedModule()
            ]).get('commands.command').execute()

            if has_event_emitter:
                app.get('events.emitter').emit('app:stop', {})

    argument_parser = ArgumentParser(add_help=False)

    class ManagedModule(AbstractModule):

        def define_context(self, context):
            return context.nested(Container({
                'commands.argument_parser': Parameter(argument_parser)
            }))

    class ManagementModule(AbstractModule):

        def define_context(self, context):
            return context.nested(Container({
                'management.command': Parameter(ManagementCommand(
                    argument_parser=argument_parser
                ))
            }))


    app = Application([
        ManagementModule()
    ])

    app.get('management.command').execute()


from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Optional, Mapping, Iterable, Any, Self


class AbstractContainer(ABC):

    @abstractmethod
    def has(self, name: str, context: Optional[AbstractContext] = None) -> bool:
        ...
    
    @abstractmethod
    def get(self, name: str, context: Optional[AbstractContext] = None) -> Any:
        ...
    
    @abstractmethod
    def resources(self, context: Optional[AbstractContext] = None) -> Mapping[str, AbstractResource]:
        ...


class AbstractResource(ABC):

    @abstractmethod
    def resolve(self, context: AbstractContext, name: str = 'anonymous'):
        ...


class AbstractContext(AbstractContainer):

    @abstractmethod
    def nested(self, container: AbstractContainer) -> AbstractContext:
        ...
    
    @abstractmethod
    def children(self) -> Iterable[AbstractContext]:
        ...
    
    @abstractmethod
    def parent(self) -> Optional[AbstractContext]:
        ...


class AbstractReference(ABC):

    @abstractmethod
    def get(self, context: AbstractContext):
        ...


class ParentReference(AbstractReference):
    pass


class AbstractRegistry[T](ABC):

    @abstractmethod
    def register(self, name: str, _: T) -> Self:
        ...
    
    @abstractmethod
    def all(self) -> Mapping[str, T]:
        ...
    




from logos_cdi.abstract import AbstractContext, AbstractContainer, AbstractResource
from logos_cdi.error import ResourceNotFoundException
from typing import Optional, Iterable, Mapping, Any
from weakref import WeakSet

class Context(AbstractContext):

    def __init__(self, container: AbstractContainer, parent: Optional[AbstractContext] = None):
        super().__init__()
        self._container = container
        self._parent = parent
        self._children = WeakSet()
    
    def has(self, name: str, context: Optional[AbstractContext] = None) -> bool:
        return self._container.has(name, context or self) or self._parent is not None and self._parent.has(name, context or self)
    
    def get(self, name, context = None) -> Any:    
        if self._container.has(name, context or self):
            return self._container.get(name, context or self)
        if self._parent is not None and self._parent.has(name, context or self):
            return self._parent.get(name, context or self)
        raise ResourceNotFoundException(name, self)
    
    def resources(self, context: Optional[AbstractContext] = None) -> Mapping[str, AbstractResource]:
        resource = {
            resource_name: resource
            for resource_name, resource in self._container.resources(context or self).items()
        }
        if self._parent is not None:
            resource.update(self._parent.resources(context or self))
        return resource
    
    def children(self) -> Iterable[AbstractContext]:
        return list(self.children)
    
    def parent(self) -> Optional[AbstractContext]:
        return self._parent
    
    def nested(self, container: AbstractContainer) -> AbstractContext:
        nested = Context(container, self)
        self._children.add(nested)
        return nested

    


from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Optional, MutableMapping, Any, Callable
from threading import Lock
from functools import wraps
from logos_cdi.abstract import AbstractContext
from logos_cdi.application import AbstractModule
from logos_cdi.container import Container
from logos_cdi.resource import Singleton, Parameter
from logos_cdi.reference import Parent, Reference, Context


class AbstractEventEmitter(ABC):

    @abstractmethod
    def emit(self, event: str, data: MutableMapping[str, Any]):
        ...

    @abstractmethod
    def on(self, event: str, handler: Callable[[MutableMapping[str, Any]], None]):
        ...

    @abstractmethod
    def once(self, event: str, handler: Callable[[MutableMapping[str, Any]], None]):
        ...


class EventEmitter(AbstractEventEmitter):
    
    def __init__(self, parent: Optional[AbstractEventEmitter], context: AbstractContext):
        super().__init__()
        self._parent = parent
        self._context = context
        self._registry = {}
        self._lock = Lock()
    
    def emit(self, event, data):
        listeners = []
        with self._lock:
            listeners = self._registry.setdefault(event, set())
            data.setdefault('contexts', []).insert(0, self._context)

        for listener in listeners:
            listener(data)
        
        if not data.get('stop_propagation', False) and self._parent:
            self._parent.emit(event, data)

    
    def on(self, event, handler):
        with self._lock:
            self._registry.setdefault(event, set()).add(handler)
    
    def once(self, event, handler):
        @wraps(handler)
        def wrapper(data):
            self._registry.setdefault(event, set()).remove(wrapper)
            return handler(data)
        self.on(event, wrapper)



class Events(AbstractModule):

    def define_context(self, context):
        return context.nested(Container({
            'events.emitter.class': Parameter('logos_cdi.event.EventEmitter'),
            'events.emitter': Singleton(
                class_path=Reference('events.emitter.class'),
                parameters={
                    'parent': Parent('events.emitter'),
                    'context': Context()
                }
            )
        }))



from logos_cdi.abstract import AbstractResource, AbstractContext, AbstractReference, AbstractRegistry, AbstractContainer
from logos_cdi.container import Container
from importlib import import_module
from typing import Any, Union, Mapping, MutableMapping, Optional
from weakref import WeakKeyDictionary
from contextvars import ContextVar, copy_context
from threading import Lock

class Parameter(AbstractResource):
    def __init__(self, value: Any):
        super().__init__()
        self.value = value
        self._cache = ContextVar('_cache', default=None)
        self._lock = Lock()
        self._locks = WeakKeyDictionary()

    def _recursive(self, value: Any, context: AbstractContext):
        if isinstance(value, list):
            return [self._recursive(item, context) for item in value]
        if isinstance(value, dict):
            return {item_key: self._recursive(item_value, context) for item_key, item_value in value.items()}
        if isinstance(value, AbstractReference):
            return value.get(context)
        return value

    def _resolve(self, context, name):
        if self._cache.get() is None:
            self._cache.set(WeakKeyDictionary())

        with self._lock:
            if context not in self._locks:
                self._locks[context] = Lock()
        
        if context in self._cache.get().keys():
            return self._cache.get()[context]

        with self._locks[context]:
            if context in self._cache.get().keys():
                return self._cache.get()[context]

            if isinstance(self.value, dict):
                self._cache.get()[context] = {}
                self._cache.get()[context].update(self._recursive(self.value, context))
            elif isinstance(self.value, list):
                self._cache.get()[context] = []
                self._cache.get()[context].extend(self._recursive(self.value, context))
            else:
                self._cache.get()[context] = self._recursive(self.value, context)

            return self._cache.get()[context]

    def resolve(self, context, name='anonymous'):
        return copy_context().run(self._resolve, context, name)

class Service(AbstractResource):
    def __init__(self, class_path: Union[str, AbstractReference], parameters: Union[Mapping[str, Any], AbstractReference]):
        self.class_path = class_path
        self.parameters = parameters
        self._cache = ContextVar('_cache', default=None)
        self._lock = Lock()
        self._locks = WeakKeyDictionary()

    def _resolve(self, context, name):
        if self._cache.get() is None:
            self._cache.set(WeakKeyDictionary())

        with self._lock:
            if context not in self._locks:
                self._locks[context] = Lock()

        if context in self._cache.get().keys():
            return self._cache.get()[context]

        with self._locks[context]:
            if context in self._cache.get().keys():
                return self._cache.get()[context]

            class_path = self.class_path if isinstance(self.class_path, str) else self.class_path.get(context)
            parameters = Parameter(self.parameters if isinstance(self.parameters, dict) else self.parameters.get(context))
            module_path, class_name = class_path.rsplit('.', 1)
            cls = getattr(import_module(module_path), class_name)
            self._cache.get()[context] = object.__new__(cls)
            self._cache.get()[context].__init__(**parameters.resolve(context, name=name))

            return self._cache.get()[context]

    def resolve(self, context, name='anonymous'):
        return copy_context().run(self._resolve, context, name)

class Singleton(AbstractResource):
    def __init__(self, class_path, parameters):
        super().__init__()
        self._service = Service(
            class_path=class_path,
            parameters=parameters
        )
        self._instances = WeakKeyDictionary()
        self._lock = Lock()
        self._locks = WeakKeyDictionary()

    def resolve(self, context, name='anonymous'):
        with self._lock:
            if context not in self._locks:
                self._locks[context] = Lock()

        with self._locks[context]:
            if context not in self._instances.keys():
                self._instances[context] = self._service.resolve(context, name)

        return self._instances[context]



from logos_cdi.abstract import AbstractContext, AbstractContainer
from logos_cdi.context import Context
from logos_cdi.container import Container
from logos_cdi.resource import Parameter
from abc import ABC, abstractmethod
from typing import Iterable


class AbstractModule(ABC):

    @abstractmethod
    def define_context(self, context: AbstractContext) -> AbstractContext:
        ...


class BaseModule(AbstractModule):

    def define_context(self, context):
        return context


class ModuleMixin:

    def define_context(self, context):
        return super().define_context(context)


class Application(AbstractContainer):

    def __init__(self, modules: Iterable[AbstractModule]):
        super().__init__()
        self._modules = modules 
        self._context = self._build_context(modules)
    
    def _build_context(self, modules: Iterable[AbstractModule]) -> AbstractContext:
        context = Context(Container({
            'application': Parameter(self),
        }))

        for module in modules:
            context = module.define_context(context)

        return context
    
    def has(self, name, context = None):
        return self._context.has(name, context)
    
    def get(self, name, context = None):
        return self._context.get(name, context)
    
    def resources(self, context = None):
        return super().resources(context)
    
    def extends(self, modules: Iterable[AbstractModule]):
        return Application([
            *self._modules,
            *modules
        ])




from logos_cdi.abstract import AbstractContainer, AbstractResource
from logos_cdi.error import ResourceNotFoundException
from typing import Mapping


class Container(AbstractContainer):

    def __init__(self, resources: Mapping[str, AbstractResource]):
        super().__init__()
        self._resources = resources

    def has(self, name, context = None):
        return name in self._resources.keys()
    
    def get(self, name, context = None):
        resource = self._resources.get(name)
        if resource is not None:
            return resource.resolve(context, name)
        raise ResourceNotFoundException(name, self)
    
    def resources(self, context):
        return {
            resource_name: resource
            for resource_name, resource in self._resources.items()
        }



from logos_cdi.abstract import AbstractReference, AbstractRegistry, AbstractContext


class Reference(AbstractReference):

    def __init__(self, resource_name: str):
        super().__init__()
        self.resource_name = resource_name
    
    def get(self, context):
        return context.get(self.resource_name)

class Parent(AbstractReference):

    def __init__(self, resource_name: str):
        super().__init__()
        self.resource_name = resource_name
    
    def get(self, context):
        parent = context.parent()
        return parent.get(self.resource_name) if parent.has(self.resource_name) and parent is not None else None

class Decorated(AbstractReference):

    def __init__(self, resource_name: str, current_context: AbstractContext, legacy: bool = False):
        super().__init__()
        self.resource_name = resource_name
        self.current_context = current_context
        self.legacy = legacy
    
    def get(self, context):
        return self.current_context.get(self.resource_name, context if not self.legacy else self.current_context)

class Context(AbstractReference):

    def get(self, context):
        return context
    

class AllRegisteredReference(AbstractReference):

    def __init__(self, registry_name: str, lazy: bool = False):
        super().__init__()
        self.registry_name = registry_name
        self.lazy = lazy
    
    def get(self, context):
        registry: AbstractRegistry[AbstractReference] = context.get(self.registry_name)
        return {
            name: reference.get(context) if not self.lazy else lambda: reference.get(context)
            for name, reference in registry.all().items()
        }


from typing import Optional, Type
from logos_cdi.abstract import AbstractRegistry
from threading import Lock


class Registry[T](AbstractRegistry[T]):

    def __init__(self, parent: Optional[AbstractRegistry], type: Optional[Type[T]]):
        super().__init__()
        self._parent = parent
        self._registry = {}
        self._type = type
        self._lock = Lock()
    
    def register(self, name, _):
        if not isinstance(_, self._type):
            raise ValueError(f'instance of {self._type} expected, {type(_)} given')
        with self._lock:
            self._registry[name] = _
        return self

    def all(self):
        all = self._parent.all() if self._parent is not None else {}
        all.update(self._registry)
        return all


